from abc import ABC, abstractmethod
from typing import Any, List


class AbstractAsyncRepository(ABC):
    def __init__(self, session: Any) -> None:
        self.session = session

    @abstractmethod
    async def add(self, obj):
        raise NotImplementedError

    @abstractmethod
    async def get(self, *identifiers: List[Any]):
        raise NotImplementedError

    async def update(self, obj):
        raise NotImplementedError

    async def delete(self, obj) -> None:
        raise NotImplementedError

    async def all(self, *identifier: List[Any]):
        raise NotImplementedError



import pytest
from asyncio import new_event_loop, set_event_loop


loop = new_event_loop()


class BaseAsyncTest:
    @pytest.fixture(scope='session')
    def event_loop(self):
        set_event_loop(loop)
        yield loop


import logging
from .logger_with_id import LoggerWithID
from datetime import datetime


class SimpleFileLogger(LoggerWithID):
    def __init__(self, name, level=logging.DEBUG, filename=...):
        super().__init__(name, level=level)

        # Create a formatter to specify the log record's format
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        
        # Create a console handler to display logs on the console
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        self.addHandler(console_handler)
        
        if filename is not None:
            if filename is ...:
                date = datetime.now().strftime('%Y-%m-%d')
                filename = f"{name}-{date}.log"

            # Create a file handler to write logs to a file
            file_handler = logging.FileHandler(filename)
            file_handler.setFormatter(formatter)
            
            self.addHandler(file_handler)



import inspect
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Type, TypeVar, Union

from .async_repository import AbstractAsyncRepository

TRepo = TypeVar("TRepo", bound=AbstractAsyncRepository)


TUOW = TypeVar("TUOW", bound=AbstractAsyncRepository)


class AbstractUOW(ABC):
    def __init__(self, repo, session: Any) -> None:
        self.repo = repo(_uow=self)
        self.session = session

    @abstractmethod
    def get_repo(self, repo_cls: Type[TRepo]) -> TRepo:
        raise NotImplementedError


@dataclass
class BaseRepoCollector(ABC):
    _uow: Union[AbstractUOW, None] = None
    _repos = None

    def __getattribute__(self, __name: str) -> Any:
        try:
            attr = super().__getattribute__(__name)
            return attr
        except AttributeError:
            if self._repos is None:
                self.init_repos()

            if self._repos is not None:
                repo = self._repos.get(__name)

                if repo is None:
                    raise
                if self._uow is None:
                    raise NotImplementedError

                return self._uow.get_repo(repo)

            raise

    def init_repos(self):
        self._repos = {}

        for name_, type_ in self.__class__.__annotations__.items():
            if not inspect.isclass(type_):
                raise NotImplementedError

            if issubclass(type_, AbstractAsyncRepository):
                self._repos[name_] = type_


class AbstractUOWFactory(ABC):
    def __init__(self, session_maker, repo: Type[BaseRepoCollector]) -> None:
        self.session_maker = session_maker
        self.repo = repo

    @abstractmethod
    def __call__(self) -> AbstractUOW:
        raise NotImplementedError
